from typing import Union

from .. import api, Elf, StrTabEditor
from ..constants import DYNAMIC_ENTRY_TAG, STR_DT, PTR_DT, ELF32, ELF64, EDITOR, SECTION

from ..core.element import Element


class DynamicEntry(api.DynamicEntry, Element):
    def __init__(self, tag: Union[DYNAMIC_ENTRY_TAG, int], un: Union[str, int]):
        self.tag = tag
        self.un = un

    @classmethod
    def size(cls, elf: Elf) -> int:
        if elf.unit == ELF32:
            return int(elf.unit.Sword) + int(elf.unit.Word)
        elif elf.unit == ELF64:
            return int(elf.unit.Xword) * 2
        else:
            raise ValueError

    @classmethod
    def from_bytes(cls, elf: Elf, b: bytes):
        assert len(b) == DynamicEntry.size(elf)

        pos = 0

        if elf.unit == ELF32:
            d_tag = int.from_bytes(b[pos:(pos := pos + int(elf.unit.Sword))], byteorder=elf.endian, signed=True)
            d_un = int.from_bytes(b[pos:(pos := pos + int(elf.unit.Word))], byteorder=elf.endian, signed=False)
        elif elf.unit == ELF64:
            d_tag = int.from_bytes(b[pos:(pos := pos + int(elf.unit.Xword))], byteorder=elf.endian, signed=True)
            d_un = int.from_bytes(b[pos:(pos := pos + int(elf.unit.Xword))], byteorder=elf.endian, signed=False)
        else:
            raise ValueError

        try:
            tag = DYNAMIC_ENTRY_TAG(d_tag)
        except ValueError:
            tag = d_tag

        dynstr_editor: StrTabEditor = elf.get_editor(EDITOR.STRTAB, SECTION.DYNSTR)

        if tag in STR_DT:
            un = dynstr_editor.get_str(d_un)
        else:
            un = d_un

        return DynamicEntry(tag, un)

    def __to_bytes(self, elf: Elf, tag_size: int, un_size: int) -> bytes:
        b = int(self.tag).to_bytes(tag_size, byteorder=elf.endian, signed=True)

        if isinstance(self.un, str):
            dynstr_editor: StrTabEditor = elf.get_editor(EDITOR.STRTAB, SECTION.DYNSTR)
            if (self.tag in STR_DT) and (off := dynstr_editor.find(self.un)) >= 0:
                b += off.to_bytes(un_size, byteorder=elf.endian, signed=True)
            else:
                raise ValueError
        else:
            b += int(self.un).to_bytes(un_size, byteorder=elf.endian, signed=True)

        assert len(b) == DynamicEntry.size(elf)

        return b

    def to_bytes(self, elf: Elf):
        if elf.unit == ELF32:
            return self.__to_bytes(elf, int(elf.unit.Sword), int(elf.unit.Word))
        elif elf.unit == ELF64:
            return self.__to_bytes(elf, int(elf.unit.Xword), int(elf.unit.Xword))
        else:
            raise ValueError

    def __str__(self):
        if hasattr(self.tag, 'name'):
            string = str(self.tag.name)
        else:
            string = hex(self.tag)
        if self.tag in PTR_DT:
            string += ', ptr: ' + hex(self.un)
        elif self.tag in STR_DT:
            string += ', val: ' + str(self.un)
        else:
            string += ', val: ' + hex(self.un)
        return string



from . import ElfHeaderEditor
from .. import api, EDITOR
from ..core import Editor
from ..elements.program_header import ProgramHeader, ProgramHeaderTable


class ProgramHeaderEditor(Editor, api.ProgramHeaderEditor):
    def read_program_header_table(self, rev_idx: int = -1) -> ProgramHeaderTable:
        elfhdr_editor: ElfHeaderEditor = self.elf.get_editor(EDITOR.ELF_HEADER)
        elfhdr = elfhdr_editor.read_elf_header(rev_idx=rev_idx)
        rev = self.elf.revisions[rev_idx]
        cache = self.elf.get_cache(rev, 'pht')

        with open(rev, 'rb') as f:
            f.seek(elfhdr.phoff)
            pht_bytes = f.read(elfhdr.phnum * elfhdr.phentsize)

        assert elfhdr.phentsize == ProgramHeader.size(self.elf)

        pht = ProgramHeaderTable()

        for i in range(elfhdr.phnum):
            ph = ProgramHeader.from_bytes(self.elf, pht_bytes[i*elfhdr.phentsize:(i+1)*elfhdr.phentsize])
            pht.append(ph)

        cache.update(pht)

        return pht

    def write_program_header_table(self, pht: ProgramHeaderTable):
        elfhdr_editor: ElfHeaderEditor = self.elf.get_editor(EDITOR.ELF_HEADER)
        elfhdr = elfhdr_editor.read_elf_header()
        rev = self.elf.get_current_revision()
        cache = self.elf.get_cache(rev, 'pht')

        with open(rev, 'r+b') as f:
            f.seek(elfhdr.phoff)
            for ph in pht:
                f.write(ph.to_bytes(self.elf))

        elfhdr.phnum = len(pht)

        elfhdr_editor.write_elf_header(elfhdr)

        cache.invalidate()



from hexdump import hexdump

from . import ElfHeaderEditor
from .. import api, SECTION, EDITOR
from ..core import Editor
from ..elements.section_header import SectionHeader, SectionHeaderTable


class SectionHeaderEditor(Editor, api.SectionHeaderEditor):
    def read_section_header_table(self, rev_idx: int = -1) -> SectionHeaderTable:
        elfhdr_editor: ElfHeaderEditor = self.elf.get_editor(EDITOR.ELF_HEADER)
        elfhdr = elfhdr_editor.read_elf_header(rev_idx=rev_idx)
        rev = self.elf.revisions[rev_idx]
        cache = self.elf.get_cache(rev, 'sht')

        if sht := cache.lookup():
            return sht

        with open(rev, 'rb') as f:
            f.seek(elfhdr.shoff)
            sht_bytes = f.read(elfhdr.shnum * elfhdr.shentsize)

        assert elfhdr.shentsize == SectionHeader.size(self.elf)

        sht = SectionHeaderTable()

        for i in range(elfhdr.shnum):
            sh = SectionHeader.from_bytes(self.elf, sht_bytes[i*elfhdr.shentsize:(i+1)*elfhdr.shentsize])
            sht.append(sh)

        cache.update(sht)

        return sht

    def read_section_header(self, section: SECTION, rev_idx: int = -1) -> SectionHeader:
        for sh in self.read_section_header_table(rev_idx=rev_idx):
            if sh.name == str(section):
                return sh

    def __get_section_header_offset_by_name(self, section: SECTION):
        elfhdr_editor: ElfHeaderEditor = self.elf.get_editor(EDITOR.ELF_HEADER)
        elfhdr = elfhdr_editor.read_elf_header()

        with open(self.elf.get_current_revision(), 'rb') as f:
            f.seek(elfhdr.shoff)
            sht_bytes = f.read(elfhdr.shnum * elfhdr.shentsize)

        assert elfhdr.shentsize == SectionHeader.size(self.elf)

        offset = -1
        for i in range(elfhdr.shnum):
            sh = SectionHeader.from_bytes(self.elf, sht_bytes[i*elfhdr.shentsize:(i+1)*elfhdr.shentsize])
            if sh.name == str(section):
                offset = elfhdr.shoff + i * elfhdr.shentsize
                break

        return offset

    def write_section_header_table(self, sht: SectionHeaderTable):
        elfhdr_editor: ElfHeaderEditor = self.elf.get_editor(EDITOR.ELF_HEADER)
        elfhdr = elfhdr_editor.read_elf_header()
        rev = self.elf.get_current_revision()
        cache = self.elf.get_cache(rev, 'sht')

        with open(rev, 'r+b') as f:
            f.seek(elfhdr.shoff)
            for sh in sht:
                f.write(sh.to_bytes(self.elf))

        elfhdr.shnum = len(sht)

        elfhdr_editor.write_elf_header(elfhdr)

        cache.invalidate()

    def write_section_header(self, section: SECTION, sh: SectionHeader):
        offset = self.__get_section_header_offset_by_name(section)
        rev = self.elf.get_current_revision()
        cache = self.elf.get_cache(rev, 'sht')

        if offset < 0:
            raise ValueError

        with open(rev, 'r+b') as f:
            f.seek(offset)
            f.write(sh.to_bytes(self.elf))

        cache.invalidate()



from .. import SECTION
from ..core import Editor
from ..elements import Symbol, SymbolTable


class SymbolEditor(Editor):
    def read_symbol_table(self, rev_idx: int = -1) -> SymbolTable:
        dynsym = self.elf.get_section(SECTION.DYNSYM)
        rev = self.elf.revisions[rev_idx]
        cache = self.elf.get_cache(rev, 'symbol')

        if st := cache.lookup():
            return st

        c = dynsym.read_content(rev_idx=rev_idx)

        st = SymbolTable()

        symbol_size = Symbol.size(self.elf)

        while c:
            symbol = Symbol.from_bytes(self.elf, c[:symbol_size])
            st.append(symbol)
            c = c[symbol_size:]

        cache.update(st)

        return st

    def write_symbol_table(self, st: SymbolTable):
        dynsym = self.elf.get_section(SECTION.DYNSYM)
        rev = self.elf.get_current_revision()
        cache = self.elf.get_cache(rev, 'symbol')

        b = bytes()

        for s in st:
            s: Symbol
            b += s.to_bytes(self.elf)

        dynsym.write_content(b)

        cache.invalidate()



from .. import api, Elf, List, Union, ELF32, ELF64, EDITOR, SECTION, StrTabEditor
from ..core import Element


class SectionHeader(Element, api.SectionHeader):
    @classmethod
    def units(cls, elf: Elf) -> List[Union[ELF32, ELF64]]:
        return [elf.unit.Word, elf.unit.Word, elf.unit.Xword, elf.unit.Addr,
                elf.unit.Off, elf.unit.Xword, elf.unit.Word, elf.unit.Word,
                elf.unit.Xword, elf.unit.Xword]

    @classmethod
    def from_bytes(cls, elf: Elf, b: bytes):
        sh_name, sh_type, sh_flags, sh_addr, sh_offset, sh_size, sh_link, sh_info, sh_addralign, sh_entsize \
            = cls.deserialize(elf, b)

        shstrtab: StrTabEditor = elf.get_editor(EDITOR.STRTAB, SECTION.SHSTRTAB, _unsafe=True)

        sh = SectionHeader()
        sh.name = shstrtab.get_str(sh_name)
        sh.type = sh_type
        sh.flags = sh_flags
        sh.addr = sh_addr
        sh.offset = sh_offset
        sh.siz = sh_size
        sh.link = sh_link
        sh.info = sh_info
        sh.addralign = sh_addralign
        sh.entsize = sh_entsize

        return sh

    def to_bytes(self, elf: Elf):
        shstrtab: StrTabEditor = elf.get_editor(EDITOR.STRTAB, SECTION.SHSTRTAB)
        return self.serialize(elf, shstrtab.find(self.name), self.type, self.flags, self.addr,
                              self.offset, self.siz, self.link, self.info,
                              self.addralign, self.entsize)

    def __str__(self):
        string = 'Section Header {'
        string += 'name: ' + str(self.name) + ', '
        string += 'type: ' + hex(self.type) + ', '
        string += 'flags: ' + hex(self.flags) + ', '
        string += 'addr: ' + hex(self.addr) + ', '
        string += 'offset: ' + hex(self.offset) + ', '
        string += 'size: ' + hex(self.siz) + ', '
        string += 'link: ' + hex(self.link) + ', '
        string += 'info: ' + hex(self.info) + ', '
        string += 'addralign: ' + hex(self.addralign) + ', '
        string += 'entsize: ' + hex(self.entsize) + '}'
        return string


class SectionHeaderTable(api.SectionHeaderTable):
    def __str__(self):
        string = '=== Section Header Table ===\n'
        for sh in self:
            string += str(sh) + '\n'
        return string



import shutil
import tempfile
from typing import List, Union

import sh

from .. import api, Elf, SectionHeaderEditor
from ..constants import SECTION, EDITOR
from ..util import readelf_hexdump_to_bytes


class Section(api.Section):
    elf: Elf
    readsection: callable

    def __init__(self, elf: Elf, section: SECTION):
        self.elf = elf
        self.name = section.value
        self.tag = section
        self.readsection = elf.readelf.bake(x=self.name)

    def read_content(self, rev_idx: int = -1) -> bytes:
        rev = self.elf.revisions[rev_idx]
        cache = self.elf.get_cache(rev, 'sec: ' + str(self.tag))

        if content := cache.lookup():
            return content

        if self.name == str(SECTION.SHSTRTAB):
            # Unfortunately, objdump cannot dump .shstrtab
            content = readelf_hexdump_to_bytes(self.readsection(rev))
            cache.update(content)
            return content

        tmpf = tempfile.mktemp(suffix=self.name, dir=self.elf.workdir.name)

        try:
            self.elf.objcopy('--dump-section', self.name + '=' + tmpf, rev)

            with open(tmpf, 'rb') as f:
                content = f.read()
        except (FileNotFoundError, sh.ErrorReturnCode):
            return readelf_hexdump_to_bytes(self.readsection(rev))

        cache.update(content)

        return content

    def write_content(self, content: bytes):
        current_rev = self.elf.get_current_revision()
        next_rev = tempfile.mktemp(dir=self.elf.workdir.name)

        if len(content) == len(self.read_content()):
            she: SectionHeaderEditor = self.elf.get_editor(EDITOR.SECTION_HEADER)
            shutil.copy2(current_rev, next_rev)
            with open(next_rev, 'r+b') as f:
                f.seek(she.read_section_header(self.tag).offset)
                f.write(content)
        else:
            tmpf = tempfile.mktemp(suffix=self.name, dir=self.elf.workdir.name)

            with open(tmpf, 'wb') as f:
                f.write(content)

            self.elf.objcopy('--update-section', self.name + '=' + tmpf,
                             current_rev, next_rev)

        self.elf.revisions.append(next_rev)

    def find(self, target: Union[bytes, str]) -> int:
        if isinstance(target, str):
            target = target.encode('ascii')

        return self.read_content().find(target)

    def __str__(self):
        return self.name



from typing import List, Union

from .e_ident import E_Ident
from ..constants import ELF_VERSION, ELF_MACHINE, ELF_TYPE, EDITOR, ELF32, ELF64
from .. import Elf, api
from ..core import Element


class ElfHeader(api.ElfHeader, Element):
    ident: E_Ident

    def __init__(self, ident, typ, machine, version, entry, phoff, shoff, flags, ehsize,
                 phentsize, phnum, shentsize, shnum, shstrndx):
        self.ident = ident
        self.typ = typ
        self.machine = machine
        self.version = version
        self.entry = entry
        self.phoff = phoff
        self.shoff = shoff
        self.flags = flags
        self.ehsize = ehsize
        self.phentsize = phentsize
        self.phnum = phnum
        self.shentsize = shentsize
        self.shnum = shnum
        self.shstrndx = shstrndx

    @classmethod
    def size(cls, elf: Elf) -> int:
        return super(ElfHeader, cls).size(elf) + E_Ident.size()

    @classmethod
    def units(cls, elf: Elf) -> List[Union[ELF32, ELF64]]:
        return [elf.unit.Half, elf.unit.Half, elf.unit.Word, elf.unit.Addr,
                elf.unit.Off, elf.unit.Off, elf.unit.Word, elf.unit.Half,
                elf.unit.Half, elf.unit.Half, elf.unit.Half, elf.unit.Half,
                elf.unit.Half]

    @classmethod
    def from_bytes(cls, elf: Elf, b: bytes):
        assert len(b) == cls.size(elf)

        ident = elf.get_editor(EDITOR.ELF_HEADER).read_e_ident()
        pos = ident.size()

        e_type, e_machine, e_version, e_entry,\
        e_phoff, e_shoff, e_flags, e_ehsize,\
        e_phentsize, e_phnum, e_shentsize, e_shnum,\
        e_shstrndx = cls.deserialize(elf, b[pos:])

        typ = ELF_TYPE(e_type)
        machine = ELF_MACHINE(e_machine)
        version = ELF_VERSION(e_version)

        return ElfHeader(ident,
                         typ, machine, version, e_entry,
                         e_phoff, e_shoff, e_flags, e_ehsize,
                         e_phentsize, e_phnum, e_shentsize, e_shnum,
                         e_shstrndx)

    def to_bytes(self, elf: Elf):
        return self.ident.to_bytes() + self.serialize(elf,
            int(self.typ), int(self.machine), int(self.version), self.entry,
            self.phoff, self.shoff, self.flags, self.ehsize,
            self.phentsize, self.phnum, self.shentsize, self.shnum,
            self.shstrndx)

    def __str__(self):
        string = 'ELF HEADER{'
        string += str(self.ident) + ', '
        string += 'type: ' + str(self.typ) + ', '
        string += 'machine: ' + str(self.machine) + ', '
        string += 'version: ' + str(self.version) + ', '
        string += 'entry: ' + hex(self.entry) + ', '
        string += 'phoff: ' + hex(self.phoff) + ', '
        string += 'shoff: ' + hex(self.shoff) + ', '
        string += 'flags: ' + hex(self.flags) + ', '
        string += 'ehsize: ' + hex(self.ehsize) + ', '
        string += 'phentsize: ' + hex(self.phentsize) + ', '
        string += 'phnum: ' + hex(self.phnum) + ', '
        string += 'shentsize: ' + hex(self.shentsize) + ', '
        string += 'shnum: ' + hex(self.shnum) + ', '
        string += 'shstrndx: ' + hex(self.shstrndx) + '}'
        return string



from typing import Union, List

from .. import ELF64, ELF32
from ..api import Elf


class Element:
    @classmethod
    def units(cls, elf: Elf) -> List[Union[ELF32, ELF64]]:
        raise NotImplementedError

    @classmethod
    def from_bytes(cls, elf: Elf, b: bytes):
        raise NotImplementedError

    def to_bytes(self, elf: Elf) -> bytes:
        raise NotImplementedError

    @classmethod
    def __size(cls, elf: Elf) -> int:
        s = 0
        for unit in cls.units(elf):
            s += int(unit)
        return s

    @classmethod
    def size(cls, elf: Elf) -> int:
        return cls.__size(elf)

    @classmethod
    def deserialize(cls, elf: Elf, b: bytes):
        results = []
        pos = 0

        assert len(b) == cls.__size(elf)

        for unit in cls.units(elf):
            signed = unit in [ELF32.Sword, ELF32.Sxword, ELF64.Sword, ELF64.Sxword]
            val = int.from_bytes(b[pos:(pos := pos + int(unit))], byteorder=elf.endian, signed=signed)
            results.append(val)

        if len(results) <= 0:
            raise ValueError
        elif len(results) == 1:
            return results[0]
        else:
            return tuple(results)

    @classmethod
    def serialize(cls, elf: Elf, *values: int) -> bytes:
        b = bytes()

        assert len(values) == len(cls.units(elf))

        for value, unit in zip(values, cls.units(elf)):
            signed = unit in [ELF32.Sword, ELF32.Sxword, ELF64.Sword, ELF64.Sxword]
            b += int(value).to_bytes(int(unit), byteorder=elf.endian, signed=signed)

        assert len(b) == cls.__size(elf)

        return b



import tempfile
from typing import List, Union, Dict, Tuple

from .constants import EDITOR, SECTION, DYNAMIC_ENTRY_TAG, ELF32, ELF64, SEGMENT_TYPE, SYMBOL_BIND, SYMBOL_TYPE, \
    SYMBOL_VISIBILITY, ELF_TYPE, ELF_MACHINE, ELF_VERSION


class ElfHeader:
    typ: Union[ELF_TYPE, int]
    machine: Union[ELF_MACHINE, int]
    version: Union[ELF_VERSION, int]
    entry: int
    phoff: int
    shoff: int
    flags: int
    ehsize: int
    phentsize: int
    phnum: int
    shentsize: hex
    shnum: int
    shstrndx: int


class Section:
    name: str

    def read_content(self, rev_idx: int = -1) -> bytes:
        raise NotImplementedError

    def write_content(self, content: bytes):
        raise NotImplementedError

    def find(self, target: Union[bytes, str]) -> int:
        raise NotImplementedError


class Editor:
    pass
    # def get_section(self, section: SECTION):
    #     raise NotImplementedError


class Elf:
    revisions: List[str]

    workdir: tempfile.TemporaryDirectory

    objcopy: callable
    objdump: callable
    readelf: callable

    unit: Union[type(ELF32), type(ELF64)]
    endian: str

    cache: Dict[str, object]

    def get_editor(self, typ: EDITOR, *args, **kwargs) -> any:
        raise NotImplementedError

    def get_section(self, section: SECTION) -> Section:
        raise NotImplementedError

    def get_current_revision(self):
        raise NotImplementedError

    def iter_objdump_sections(self):
        class S:
            idx: int
            name: str
            size: int
            vma: int
            lma: int
            file_off: int
            align: int
        raise NotImplementedError

    def auto_adjust(self):
        raise NotImplementedError

    def write(self, path: str, auto_adjust: bool = True, mkdirs=True):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    class Cache:
        key: str
        kv: dict

        def __init__(self, elf, key: str):
            self.key = key
            self.kv = elf.cache

        def lookup(self):
            try:
                return self.kv[self.key]
            except KeyError:
                return None

        def update(self, b: object):
            self.kv[self.key] = b

        def invalidate(self):
            if (key := self.key) in self.kv.keys():
                self.kv.pop(key)

    def get_cache(self, rev: str, key: str):
        return self.Cache(self, 'rev: ' + rev + ', extra: ' + key)


class ElfHeaderEditor:
    def read_elf_header(self, rev_idx: int = -1) -> ElfHeader:
        raise NotImplementedError

    def write_elf_header(self, elf_header: ElfHeader):
        raise NotImplementedError


class ProgramHeader:
    type: Union[SEGMENT_TYPE, int]
    offset: int
    vaddr: int
    paddr: int
    filesz: int
    memsz: int
    flags: int
    align: int


class ProgramHeaderTable(List[ProgramHeader]):
    pass


class ProgramHeaderEditor:
    def read_program_header_table(self, rev_idx: int = -1) -> ProgramHeaderTable:
        raise NotImplementedError

    def write_program_header_table(self, sht: ProgramHeaderTable):
        raise NotImplementedError


class SectionHeader:
    name: str
    type: int
    flags: int
    addr: int
    offset: int
    siz: int
    link: int
    info: int
    addralign: int
    entsize: int


class SectionHeaderTable(List[SectionHeader]):
    pass


class SectionHeaderEditor:
    def read_section_header_table(self, rev_idx: int = -1) -> SectionHeaderTable:
        raise NotImplementedError

    def read_section_header(self, section: SECTION, rev_idx: int = -1) -> SectionHeader:
        raise NotImplementedError

    def write_section_header_table(self, sht: SectionHeaderTable):
        raise NotImplementedError

    def write_section_header(self, section: SECTION, sh: SectionHeader):
        raise NotImplementedError


class Version:
    soname: str
    name: str

    def is_local(self):
        raise NotImplementedError

class Verdef:
    def update(self, new_version: str, ndx: int):
        raise NotImplementedError

    def get_ndx(self):
        raise NotImplementedError


class Verdaux:
    pass


class Verneed:
    pass


class Vernaux:
    other: int
    name: str


class VersionTable(List[Version]):
    pass


class VerdefTable:
    def __init__(self):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError

    def add_entry(self, vda_name: str) -> Verdef:
        raise NotImplementedError


class VerneedTable:
    def __init__(self):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError

    def add_entry(self, vn_file: str, vna_name: str):
        raise NotImplementedError


class SymbolVersionEditor(Editor):
    def read_versions(self, rev_idx: int = -1) -> VersionTable:
        raise NotImplementedError

    def read_version_definition(self, rev_idx: int = -1) -> VerdefTable:
        raise NotImplementedError

    def read_version_requirement(self, rev_idx: int = -1) -> VerneedTable:
        raise NotImplementedError

    def write(self, version_table: VersionTable = None, verdef_table: VerdefTable = None,
              verneed_table: VerneedTable = None):
        raise NotImplementedError

    def get_vername_soname_by_version(self, idx: int) -> Tuple[str, str]:
        raise NotImplementedError


class DynamicEntry:
    tag: Union[DYNAMIC_ENTRY_TAG, int]
    un: Union[int, str, None]


class DynamicEntryEditor(Editor):
    def read_dynamic_entries(self, rev_idx: int = -1) -> List[DynamicEntry]:
        raise NotImplementedError

    def write_dynamic_entries(self, dynlist):
        raise NotImplementedError

    def read_soname(self, rev_idx: int = -1):
        raise NotImplementedError


class StrTabEditor(Editor):
    def append(self, string: str):
        raise NotImplementedError

    def find(self, string: str) -> int:
        raise NotImplementedError

    def has(self, string: str) -> bool:
        raise NotImplementedError

    def get_str(self, pos: int) -> str:
        raise NotImplementedError


class Symbol:
    name: str
    value: int
    siz: int
    bind: SYMBOL_BIND
    typ: SYMBOL_TYPE
    other: SYMBOL_VISIBILITY
    shndx: int

    tag: SECTION

    def is_defined(self):
        raise NotImplementedError

    def is_needed(self):
        raise NotImplementedError


class SymbolTable(List[Symbol]):
    def defined_symbols(self):
        raise NotImplementedError

    def needed_symbols(self):
        raise NotImplementedError


class SymbolEditor:
    def read_symbol_table(self, rev_idx: int = -1) -> SymbolTable:
        raise NotImplementedError

    def write_symbol_table(self, st: SymbolTable):
        raise NotImplementedError


class GnuHash:
    nbuckets: int
    symndx: int
    maskwords: int
    shift2: int
    bloom_filter: List[int]
    buckets: List[int]
    values: List[int]


class NotAnElfError(Exception):
    pass


def parse(path: str, toolchain_path: str = None, prefix: str = '') -> Elf:
    from .core import Elf
    return Elf(path,
               toolchain_path=toolchain_path,
               prefix=prefix)



def gnu_hash_binutils_bfd_elf(string: str):
    """ toolchain/binutils/binutils-2.27/bfd/elf.c

    /* Standard ELF hash function.  Do not change this function; you will
       cause invalid hash tables to be generated.  */

    unsigned long
    bfd_elf_hash (const char *namearg)
    {
      const unsigned char *name = (const unsigned char *) namearg;
      unsigned long h = 0;
      unsigned long g;
      int ch;

      while ((ch = *name++) != '\0')
        {
          h = (h << 4) + ch;
          if ((g = (h & 0xf0000000)) != 0)
        {
          h ^= g >> 24;
          /* The ELF ABI says `h &= ~g', but this is equivalent in
             this case and on some machines one insn instead of two.  */
          h ^= g;
        }
        }
      return h & 0xffffffff;
    }
    """
    h = 0
    for ch in list(string.encode(encoding='ascii')):
        h = (h << 4) + ch
        if (g := (h & 0xf0000000)) != 0:
            h ^= g >> 24
            h ^= g

    return h & 0xffffffff


def gnu_hash_bionic_linker_linker_sofinfo(string: str):
    """ bionic/linker/linker_soinfo.cpp

    uint32_t calculate_elf_hash(const char* name) {
      const uint8_t* name_bytes = reinterpret_cast<const uint8_t*>(name);
      uint32_t h = 0, g;

      while (*name_bytes) {
        h = (h << 4) + *name_bytes++;
        g = h & 0xf0000000;
        h ^= g;
        h ^= g >> 24;
      }

      return h;
    }
    """
    h = 0
    for ch in list(string.encode(encoding='ascii')):
        h = ((h << 4) + ch) & 0xffffffff
        g = h & 0xf0000000
        h ^= g
        h ^= g >> 24

    return h


def gnu_hash(string: str):
    return gnu_hash_bionic_linker_linker_sofinfo(string)
    # return gnu_hash_binutils_bfd_elf(string)


def readelf_hexdump_to_bytes(hexdump: str) -> bytes:
    b = bytes()
    for line in [e for e in hexdump.splitlines()][2:]:
        if len(line) == 0:
            continue
        if (end := line.find('  ', 2)) > 0:
            line = line[:end]
        addr, *words = line.split()
        try:
            for word in words[:4]:
                b += int(word, 16).to_bytes(int(len(word)/2), byteorder='big')
        except ValueError:
            raise AssertionError('error: Failed at parsing hexdump')

    return b


def unpack_bytes_to_ints(b: bytes, segment_size: int, byteorder: str, signed=False):
    assert (len(b) % segment_size) == 0
    return [int.from_bytes(b[i:i+segment_size], byteorder, signed=signed) for i in range(0, len(b), segment_size)]


import os
import re
import tempfile
from typing import List, Tuple

import capstone
import sh
import shutil

from hexdump import hexdump

from .. import api, Editor, DynamicEntryEditor, ElfHeader, SectionHeader, SectionHeaderTable, ProgramHeaderTable, \
    ProgramHeader, SectionHeaderEditor, ProgramHeaderEditor, SymbolTable, ElfHeaderEditor, SymbolEditor, Symbol, \
    SymbolVersionEditor, NotAnElfError
from .section import Section
from .segment import Segment
from ..constants import EDITOR, SECTION, DYNAMIC_ENTRY_TAG, ELF32, ELF64, ELF_CLASS, SEGMENT_TYPE
from ..util import unpack_bytes_to_ints


class Elf(api.Elf):

    def get_editor(self, typ: EDITOR, *args, **kwargs) -> any:
        from ..editors import SymbolVersionEditor, DynamicEntryEditor, StrTabEditor, ElfHeaderEditor, \
            SectionHeaderEditor, ProgramHeaderEditor, SymbolEditor

        if typ is EDITOR.SYMBOL_VERSION:
            editor = SymbolVersionEditor(self)
        elif typ is EDITOR.DYNAMIC_ENTRY:
            editor = DynamicEntryEditor(self)
        elif typ is EDITOR.STRTAB:
            editor = StrTabEditor(self, *args, **kwargs)
        elif typ is EDITOR.ELF_HEADER:
            editor = ElfHeaderEditor(self)
        elif typ is EDITOR.SECTION_HEADER:
            editor = SectionHeaderEditor(self)
        elif typ is EDITOR.PROGRAM_HEADER:
            editor = ProgramHeaderEditor(self)
        elif typ is EDITOR.SYMBOL:
            editor = SymbolEditor(self)
        else:
            raise TypeError

        return editor

    def get_tmpdir(self):
        if os.path.isdir('/dev/shm'):
            return '/dev/shm'
        return None

    def __init__(self, path: str,
                 toolchain_path: str = None,
                 prefix: str = ''):
        self.revisions = [path]
        self.lock = {}
        self.workdir = tempfile.TemporaryDirectory(dir=self.get_tmpdir(), prefix='woodelf-')

        if toolchain_path:
            self.objcopy = sh.Command(prefix + 'objcopy', search_paths=[toolchain_path]).bake('--pure')
            self.objdump = sh.Command(prefix + 'objdump', search_paths=[toolchain_path])
            self.readelf = sh.Command(prefix + 'readelf', search_paths=[toolchain_path])
        else:
            self.objcopy = sh.Command(prefix + 'objcopy').bake('--pure')
            self.objdump = sh.Command(prefix + 'objdump')
            self.readelf = sh.Command(prefix + 'readelf')

        self.cache = {}

        e_ident = self.get_editor(EDITOR.ELF_HEADER).read_e_ident()

        if e_ident.cls == ELF_CLASS.CLASS32:
            self.unit = ELF32
        elif e_ident.cls == ELF_CLASS.CLASS64:
            self.unit = ELF64
        else:
            raise ValueError('Invalid ELF class type')

        self.endian = str(e_ident.data)

    def get_current_revision(self):
        return self.revisions[-1]

    def get_section(self, section: SECTION, _unsafe=False) -> Section:
        from woodelf.core.section import Section

        if not _unsafe:
            she: SectionHeaderEditor = self.get_editor(EDITOR.SECTION_HEADER)
            if not she.read_section_header(section):
                raise TypeError('Section ' + str(section) + ' does not exist')

        return Section(self, section)

    def get_segment(self, idx: int) -> Segment:
        return Segment(self, )

    def num_segment(self) -> int:
        pass

    def __auto_adjust_dyn_ent_ptr(self):
        dyn_editor: DynamicEntryEditor = self.get_editor(EDITOR.DYNAMIC_ENTRY)
        dyn_entries = dyn_editor.read_dynamic_entries()

        def __update_dyn_ent(tag: DYNAMIC_ENTRY_TAG, un: int):
            for ent in filter(lambda e: e.tag == tag, dyn_entries):
                ent.un = un

        for s in self.iter_objdump_sections():
            try:
                tag: SECTION = SECTION(s.name)
            except ValueError:
                # skip unknown section
                continue

            if tag == SECTION.GOT_PLT:
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_PLTGOT, s.vma)
            elif tag == SECTION.RELA_PLT:
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_JMPREL, s.vma)
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_PLTRELSZ, s.size)
            elif tag == SECTION.RELA_DYN:
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_RELA, s.vma)
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_RELASZ, s.size)
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_ANDROID_RELA, s.vma)
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_ANDROID_RELASZ, s.size)
            elif tag == SECTION.DYNSYM:
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_SYMTAB, s.vma)
            elif tag == SECTION.DYNSTR:
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_STRTAB, s.vma)
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_STRSZ, s.size)
            elif tag == SECTION.GNU_HASH:
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_GNU_HASH, s.vma)
            elif tag == SECTION.INIT_ARRAY:
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_INIT_ARRAY, s.vma)
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_INIT_ARRAYSZ, s.size)
            elif tag == SECTION.FINI_ARRAY:
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_FINI_ARRAY, s.vma)
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_FINI_ARRAYSZ, s.size)
            elif tag == SECTION.GNU_VERSION:
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_VERSYM, s.vma)
            elif tag == SECTION.GNU_VERSION_D:
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_VERDEF, s.vma)
            elif tag == SECTION.GNU_VERSION_R:
                __update_dyn_ent(DYNAMIC_ENTRY_TAG.DT_VERNEED, s.vma)
            else:
                pass

        dyn_editor.write_dynamic_entries(dyn_entries)

    def __auto_adjust_section_vma(self):
        current_rev = self.get_current_revision()
        next_rev = tempfile.mktemp(dir=self.workdir.name)

        args = []
        after_eh_frame_hdr = False
        for s in self.iter_objdump_sections():
            if not after_eh_frame_hdr:
                if s.name == '.eh_frame_hdr':
                    after_eh_frame_hdr = True
                args.extend(['--change-section-address', s.name + '=' + str(s.lma)])
            # elif s.lma != 0:
            #     args.extend(['--change-section-address', s.name + '+' + str(s.file_off + 0x2000 - s.lma)])
            #     args.extend(['--change-section-vma', s.name + '=' + str(s.file_off)])
            #     args.extend(['--change-section-lma', s.name + '=' + str(s.file_off)])

        self.objcopy(*args, current_rev, next_rev)

        self.revisions.append(next_rev)

    def __auto_adjust_program_header(self):
        ph_editor: ProgramHeaderEditor = self.get_editor(EDITOR.PROGRAM_HEADER)
        elfhdr_editor = self.get_editor(EDITOR.ELF_HEADER)
        e = elfhdr_editor.read_elf_header()
        orig_progh = ph_editor.read_program_header_table(rev_idx=0)
        new_progh = ph_editor.read_program_header_table()

        for orig, new in zip(orig_progh, new_progh):
            orig: ProgramHeader
            new: ProgramHeader

            if orig.type == SEGMENT_TYPE.RELRO:
                sh_editor: SectionHeaderEditor = self.get_editor(EDITOR.SECTION_HEADER)
                fini_array = sh_editor.read_section_header(SECTION.FINI_ARRAY)
                data = sh_editor.read_section_header(SECTION.DATA)
                orig.offset = fini_array.offset
                orig.vaddr = fini_array.addr
                orig.paddr = fini_array.addr
                orig.filesz = data.offset - fini_array.offset
                orig.memsz = data.addr - fini_array.addr
                continue
            elif orig.type == SEGMENT_TYPE.PHDR:
                orig.offset = new.offset
                orig.vaddr = new.paddr  # paddr and vaddr must be the same
                orig.paddr = new.paddr
                # 200903 dhkim: We assume filesz and memsz are always the same (no compression)
                orig.filesz = len(orig_progh) * e.phentsize
                orig.memsz = len(orig_progh) * e.phentsize
                orig.align = new.align
                continue
            elif orig.type == SEGMENT_TYPE.STACK:
                orig.offset = new.offset
                orig.vaddr = new.paddr  # paddr and vaddr must be the same
                orig.paddr = new.paddr
                orig.filesz = new.filesz
                orig.memsz = new.memsz
                continue

            # default behavior
            if not orig or not new:
                continue

            assert orig.type == new.type

            orig.offset = new.offset
            orig.vaddr = new.paddr  # paddr and vaddr must be the same
            orig.paddr = new.paddr
            orig.filesz = new.filesz
            orig.memsz = new.memsz
            orig.align = new.align

        e.phnum = len(orig_progh)

        elfhdr_editor.write_elf_header(e)

        ph_editor.write_program_header_table(orig_progh)

    def __auto_adjust_elf_header(self):
        elfh_editor: ElfHeaderEditor = self.get_editor(EDITOR.ELF_HEADER)
        orig_elfh = elfh_editor.read_elf_header(rev_idx=0)

        elfh = elfh_editor.read_elf_header()

        elfh.flags = orig_elfh.flags

        elfh_editor.write_elf_header(elfh)

    def __auto_adjust_symbol_values(self):
        st_editor: SymbolEditor = self.get_editor(EDITOR.SYMBOL)
        sh_editor: SectionHeaderEditor = self.get_editor(EDITOR.SECTION_HEADER)

        orig_st = st_editor.read_symbol_table(rev_idx=0)
        new_st = st_editor.read_symbol_table()

        orig_sht = sh_editor.read_section_header_table(rev_idx=0)
        new_sht = sh_editor.read_section_header_table()

        assert len(orig_st) == len(new_st)

        for orig_s, new_s in zip(orig_st, new_st):
            orig_s: Symbol
            new_s: Symbol

            if not new_s.is_defined():
                continue

            assert orig_s.shndx == new_s.shndx

            offset = new_sht[new_s.shndx].offset - orig_sht[orig_s.shndx].offset

            new_s.value = orig_s.value + offset

        st_editor.write_symbol_table(new_st)

    class __AddrTranslator:
        updated_sh_pairs: List[Tuple[SectionHeader, SectionHeader]]

        def __init__(self, elf: api.Elf):
            sh_editor: SectionHeaderEditor = elf.get_editor(EDITOR.SECTION_HEADER)

            orig_sht = sh_editor.read_section_header_table(rev_idx=0)
            new_sht = sh_editor.read_section_header_table()

            self.updated_sh_pairs = []
            for orig_sh, new_sh in zip(orig_sht, new_sht):
                if orig_sh.addr != new_sh.addr:
                    self.updated_sh_pairs.append((orig_sh, new_sh))
                    
        def __to_new(self, addr: int, reverse: bool = False):
            for orig_sh, new_sh in self.updated_sh_pairs:
                sh1: SectionHeader
                sh2: SectionHeader

                if reverse:
                    sh1 = new_sh
                    sh2 = orig_sh
                else:
                    sh1 = orig_sh
                    sh2 = new_sh

                if not (sh1.addr <= addr <= sh1.addr + sh1.siz):
                    continue

                new_addr = addr + sh2.addr - sh1.addr

                return new_addr

            return addr

        def to_new(self, orig_addr: int) -> int:
            return self.__to_new(orig_addr, reverse=False)

        def to_orig(self, new_addr: int) -> int:
            return self.__to_new(new_addr, reverse=True)

    def __auto_adjust_addrs_by_heuristic(self):
        trans = self.__AddrTranslator(self)

        # for tag in [SECTION.INIT_ARRAY, SECTION.FINI_ARRAY, SECTION.DATA, SECTION.RODATA,
        #             SECTION.RELA_DYN, SECTION.RELA_PLT, SECTION.TEXT, SECTION.PLT,
        #             SECTION.DATA_REL_RO, SECTION.GNU_HASH]:

        # https://stackoverflow.com/a/7031644
        for tag in [SECTION.INIT_ARRAY, # .init_array and .fini_array are pointer to instructions
                    SECTION.FINI_ARRAY,
                    SECTION.DATA, # .data and .rodata are app specific memory. we heuristically replace pointers
                    SECTION.RODATA,
                    SECTION.DATA_REL_RO,
                    SECTION.RELA_DYN, # I don't know why but i could find addresses here
                    # SECTION.PLT,
                    # SECTION.GOT_PLT
                    ]:
            try:
                section = self.get_section(tag)
            except TypeError:
                continue

            orig_contents = section.read_content(rev_idx=0)

            if (len(orig_contents) % int(self.unit.Addr)) != 0:
                # dee: DynamicEntryEditor = self.get_editor(EDITOR.DYNAMIC_ENTRY)
                # print(self, dee.read_soname())
                # print(section)
                # hexdump(orig_contents)
                continue

            new_contents = section.read_content()

            orig_addrs = unpack_bytes_to_ints(orig_contents, int(self.unit.Addr), self.endian, signed=False)

            contents_out = bytearray(new_contents)

            for i, orig_addr in enumerate(orig_addrs):
                _new_addr = trans.to_new(orig_addr)

                contents_out[i * int(self.unit.Addr):(i + 1) * int(self.unit.Addr)] \
                    = _new_addr.to_bytes(int(self.unit.Addr), self.endian, signed=False)
                # contents_out[i:i + int(self.unit.Addr)] \
                #     = _new_addr.to_bytes(int(self.unit.Addr), self.endian, signed=False)

            section.write_content(bytes(contents_out))

    def __auto_adjust_section_gap_inner(self):
        reference_section = SECTION.EH_FRAME_HDR
        target_sections = [SECTION.RELA_DYN, SECTION.RELA_PLT, SECTION.PLT, SECTION.TEXT, SECTION.RODATA, SECTION.EH_FRAME]

        sh_editor: SectionHeaderEditor = self.get_editor(EDITOR.SECTION_HEADER)

        orig_sht = sh_editor.read_section_header_table(rev_idx=0)
        new_sht = sh_editor.read_section_header_table()

        interested_names = [e for e in map(str, target_sections)]

        filter_interested = lambda sht: sorted(filter(lambda e: e.name in interested_names, sht),
                                               key=lambda e: e.offset, reverse=True)

        orig_refsh = sh_editor.read_section_header(reference_section, rev_idx=0)
        new_refsh = sh_editor.read_section_header(reference_section)

        for orig_sh, new_sh in zip(filter_interested(orig_sht), filter_interested(new_sht)):
            orig_sh: SectionHeader
            new_sh: SectionHeader

            assert orig_sh.name == new_sh.name

            orig_gap = orig_refsh.offset - orig_sh.offset
            new_gap = new_refsh.offset - new_sh.offset

            assert orig_gap > 0 and new_gap > 0

            if orig_gap == new_gap:
                continue

            assert orig_gap > new_gap

            sec = self.get_section(SECTION(orig_sh.name))
            content = sec.read_content()
            content += b'\0' * (orig_gap - new_gap)
            sec.write_content(content)

            return False
        return True

    def __auto_adjust_section_gap_upperhalf(self):
        while not self.__auto_adjust_section_gap_inner():
            pass

    def __auto_adjust_section_gap_lowerhalf(self):
        target_sections = [SECTION.RELA_DYN, SECTION.RELA_PLT, SECTION.PLT, SECTION.TEXT, SECTION.RODATA, SECTION.EH_FRAME]
        sh_editor: SectionHeaderEditor = self.get_editor(EDITOR.SECTION_HEADER)

        for section in target_sections:
            orig_sh = sh_editor.read_section_header(section, rev_idx=0)
            new_sh = sh_editor.read_section_header(section)
            if not new_sh:
                continue
            new_sh.siz = orig_sh.siz
            sh_editor.write_section_header(section, new_sh)

    def __auto_adjust_code_x86_64(self, addr: int, contents: bytes, trans: __AddrTranslator) -> bytes:
        # https://reverseengineering.stackexchange.com/a/22225
        # http://www.capstone-engine.org/lang_python.html

        out = bytes()
        oip = trans.to_orig(addr)
        nip = addr # instruction pointer

        md = capstone.Cs(capstone.CS_ARCH_X86, capstone.CS_MODE_64)
        md.detail = True

        for insn in md.disasm(contents, addr):
            # ip points the end of an instruction, not the beginning
            oip += len(insn.bytes)
            nip += len(insn.bytes)
            insn: capstone.CsInsn

            # desc = "0x%x:\t%s\t%s\t%s" % (insn.address, hexdump(insn.bytes, result='return'), insn.mnemonic, insn.op_str)

            b = insn.bytes

            for operand in insn.operands:
                if operand.type == capstone.x86.X86_OP_MEM \
                        and operand.value.mem.base == capstone.x86.X86_REG_RIP \
                        and (disp := operand.value.mem.disp) != 0:
                    optr = oip + disp
                    nptr = nip + disp
                    corret_ptr = trans.to_new(optr)

                    if nptr == corret_ptr:
                        continue

                    nb = (disp).to_bytes(4, self.endian, signed=False)
                    cb = (corret_ptr - nip).to_bytes(4, self.endian, signed=False)

                    b = b.replace(nb, cb)
                # if operand.type == capstone.x86.X86_OP_IMM

            out += b
        return out

    def __auto_adjust_code(self):
        trans = self.__AddrTranslator(self)

        sh_editor: SectionHeaderEditor = self.get_editor(EDITOR.SECTION_HEADER)

        for section in [SECTION.PLT, SECTION.TEXT]:
            sh = sh_editor.read_section_header(section)

            s = self.get_section(section)
            content = s.read_content()

            content = self.__auto_adjust_code_x86_64(sh.addr, content, trans)

            s.write_content(content)

    def iter_objdump_sections(self):
        class S:
            idx: int
            name: str
            size: int
            vma: int
            lma: int
            file_off: int
            align: int

        lines: str = self.objdump('-h', self.get_current_revision())

        for line in lines.splitlines():
            try:
                idx, name, size, vma, lma, file_off, align = line.split(maxsplit=7)
            except ValueError:
                continue

            s = S()
            try:
                s.idx = int(idx)
                s.name = name
                s.size = int(size, 16)
                s.vma = int(vma, 16)
                s.lma = int(lma, 16)
                s.file_off = int(file_off, 16)
                s.align = int(eval(align))
            except ValueError or SyntaxError:
                continue

            yield s

    def auto_adjust(self):
        # call any adjust functions that potentially change section offset
        self.__auto_adjust_section_gap_upperhalf()

        # call any adjust functions that potentially change section address
        self.__auto_adjust_section_vma()

        # call any adjust functions that
        self.__auto_adjust_dyn_ent_ptr()
        self.__auto_adjust_program_header()
        # self.__auto_adjust_elf_header()
        self.__auto_adjust_symbol_values()
        self.__auto_adjust_addrs_by_heuristic()
        self.__auto_adjust_code()

        self.__auto_adjust_section_gap_lowerhalf()

    def write(self, path: str, auto_adjust: bool = True, mkdirs=True):
        if auto_adjust:
            self.auto_adjust()
        if mkdirs:
            os.makedirs(os.path.dirname(path), exist_ok=True)
        shutil.copy2(self.get_current_revision(), path)

    def close(self):
        self.workdir.cleanup()



from typing import Union, List, Tuple

from .. import api, gnu_hash
from ..core import Editor, Elf, Section
from ..elements import Verdef, Verdaux, Verneed, Vernaux, DynamicEntry, VersionTable, VerdefTable, VerneedTable, Version
from ..constants import SECTION, EDITOR, DYNAMIC_ENTRY_TAG

from .dynamic_entry_editor import DynamicEntryEditor
from .strtab_editor import StrTabEditor


class SymbolVersionEditor(Editor, api.SymbolVersionEditor):
    version: Section
    version_d: Section
    version_r: Section

    dynent_editor: DynamicEntryEditor
    dynstr_editor: StrTabEditor

    def __init__(self, elf: Elf):
        super().__init__(elf)

        self.version = elf.get_section(SECTION.GNU_VERSION)
        self.version_d = elf.get_section(SECTION.GNU_VERSION_D)
        self.version_r = elf.get_section(SECTION.GNU_VERSION_R)

        self.dynent_editor = self.elf.get_editor(EDITOR.DYNAMIC_ENTRY)
        self.dynstr_editor = self.elf.get_editor(EDITOR.STRTAB, SECTION.DYNSTR)

    def read_versions(self, rev_idx: int = -1) -> VersionTable:
        version_table = VersionTable()

        c = self.version.read_content(rev_idx=rev_idx)

        while c:
            ver_bytes = c[0:Version.size(self.elf)]
            c = c[Version.size(self.elf):]
            ver = Version.from_bytes(self.elf, ver_bytes)
            version_table.append(ver)

        return version_table

    def read_version_definition(self, rev_idx: int = -1) -> VerdefTable:
        verdef_table = VerdefTable()

        c = self.version_d.read_content(rev_idx=rev_idx)

        while c:
            verdef_bytes = c[0:Verdef.size(self.elf)]
            c = c[Verdef.size(self.elf):]
            verdef = Verdef.from_bytes(self.elf, verdef_bytes)
            for i in range(verdef.cnt):
                verdaux_bytes = c[0:Verdaux.size(self.elf)]
                c = c[Verdaux.size(self.elf):]
                verdaux = Verdaux.from_bytes(self.elf, verdaux_bytes)
                verdef.append_veraux(verdaux)
            verdef_table.append(verdef)

        return verdef_table

    def read_version_requirement(self, rev_idx: int = -1) -> VerneedTable:
        verneed_table = VerneedTable()

        c = self.version_r.read_content(rev_idx=rev_idx)

        while c:
            verneed_bytes = c[0:Verneed.size(self.elf)]
            c = c[Verneed.size(self.elf):]
            verneed = Verneed.from_bytes(self.elf, verneed_bytes)
            for i in range(verneed.cnt):
                vernaux_bytes = c[0:Vernaux.size(self.elf)]
                c = c[Vernaux.size(self.elf):]
                vernaux = Vernaux.from_bytes(self.elf, vernaux_bytes)
                verneed.append_veraux(vernaux)
            verneed_table.append(verneed)

        return verneed_table

    def __adjust_DT_VERDEFNUM(self, verdef_table: VerdefTable):

        dynlist: List[DynamicEntry] = self.dynent_editor.read_dynamic_entries()

        for dyn in filter(lambda e: e.tag == DYNAMIC_ENTRY_TAG.DT_VERDEFNUM, dynlist):
            dyn.un = len(verdef_table)

        self.dynent_editor.write_dynamic_entries(dynlist)

    def __adjust_DT_VERNEEDNUM(self, verneed_table: VerneedTable):
        dynlist: List[DynamicEntry] = self.dynent_editor.read_dynamic_entries()

        for dyn in filter(lambda e: e.tag == DYNAMIC_ENTRY_TAG.DT_VERNEEDNUM, dynlist):
            dyn.un = len(verneed_table)

        self.dynent_editor.write_dynamic_entries(dynlist)

    def __adjust_VERSION_D_section_info(self, verdef_table: VerdefTable):
        sheditor = self.elf.get_editor(EDITOR.SECTION_HEADER)
        sh = sheditor.read_section_header(SECTION.GNU_VERSION_D)

        sh.info = len(verdef_table)

        sheditor.write_section_header(SECTION.GNU_VERSION_D, sh)

    def __adjust_VERSION_R_section_info(self, verneed_table: VerneedTable):
        sheditor = self.elf.get_editor(EDITOR.SECTION_HEADER)
        sh = sheditor.read_section_header(SECTION.GNU_VERSION_R)

        sh.info = len(verneed_table)

        sheditor.write_section_header(SECTION.GNU_VERSION_R, sh)
        
    def __write_versions(self, version_table: VersionTable):
        assert len(version_table) == len(self.read_versions())

        b = bytes()
        for ver in version_table:
            ver: Version
            b += ver.to_bytes(self.elf)

        self.version.write_content(b)

    def __write_version_definition(self, verdef_table: VerdefTable):
        b = bytes()
        for verdef in verdef_table:
            verdef: Verdef
            b += verdef.to_bytes(self.elf)
            if not verdef.veraux_table:
                continue
            for verdaux in verdef.veraux_table:
                if not self.dynstr_editor.has(verdaux.name):
                    self.dynstr_editor.append(verdaux.name)
                b += verdaux.to_bytes(self.elf)

        self.version_d.write_content(b)

        self.__adjust_DT_VERDEFNUM(verdef_table)
        self.__adjust_VERSION_D_section_info(verdef_table)

    def __write_version_requirement(self, verneed_table: VerneedTable):
        vna_other = len(self.read_version_definition())
        b = bytes()
        for verneed in verneed_table:
            verneed: Verneed
            b += verneed.to_bytes(self.elf)
            if not verneed.veraux_table:
                continue
            for vernaux in verneed.veraux_table:
                if not self.dynstr_editor.has(vernaux.name):
                    self.dynstr_editor.append(vernaux.name)
                vernaux.other = (vna_other := vna_other + 1)

                b += vernaux.to_bytes(self.elf)

        self.version_r.write_content(b)

        self.__adjust_DT_VERNEEDNUM(verneed_table)
        self.__adjust_VERSION_R_section_info(verneed_table)

    def write(self, version_table: VersionTable = None, verdef_table: VerdefTable = None,
              verneed_table: VerneedTable = None):

        if (version_table is None) and (verdef_table is None) and (verneed_table is None):
            return

        if version_table is None:
            version_table = self.read_versions()
        if verdef_table is None:
            verdef_table = self.read_version_definition()
        if verneed_table is None:
            verneed_table = self.read_version_requirement()

        # The order of writing definitions, requirements, and versions is important
        self.__write_version_definition(verdef_table)
        self.__write_version_requirement(verneed_table)
        self.__write_versions(version_table)

    def __get_version_from_verdeftab(self, vername: str) -> int:
        verdeftab = self.read_version_definition()
        for verdef in verdeftab:
            verdef: Verdef
            if verdef.hash == gnu_hash(vername):
                return verdef.ndx
        return -1

    def __get_version_from_verneedtab(self, vername: str, soname: str = None) -> int:
        # print(self.elf.get_editor(EDITOR.DYNAMIC_ENTRY).read_soname())
        # print(self)
        # print('__get_version_from_verneedtab')
        verneedtab = self.read_version_requirement()
        for verneed in verneedtab:
            verneed: Verneed
            # print(verneed)
            if soname and (verneed.file != soname):
                continue
            if not verneed.veraux_table:
                continue
            for vernaux in verneed.veraux_table:
                vernaux: Vernaux
                # print(vernaux)
                if vernaux.name == vername:
                    return vernaux.other
        return -1

    def get_version_by_name(self, vername: str, soname: str = None) -> int:
        dynent_edit: DynamicEntryEditor = self.elf.get_editor(EDITOR.DYNAMIC_ENTRY)
        if not soname:
            if (ver := self.__get_version_from_verdeftab(vername)) < 0:
                ver = self.__get_version_from_verneedtab(vername)
        elif soname == dynent_edit.read_soname():
            ver = self.__get_version_from_verdeftab(vername)
        else:
            ver = self.__get_version_from_verneedtab(vername, soname)
        if ver < 0:
            errstr = 'No proper version associated with version ' + vername
            if soname:
                errstr += ' from ' + soname
            raise KeyError(errstr)
        return ver

    def get_vername_soname_by_version(self, idx: int) -> Tuple[str, str]:
        if idx == 0:
            raise ValueError('Local symbols (indicated by zero version number) does not have associated vername and '
                             'soname')

        dynent_editor: DynamicEntryEditor = self.elf.get_editor(EDITOR.DYNAMIC_ENTRY)

        for verdef in self.read_version_definition():
            verdef: Verdef
            if verdef.get_ndx() == idx:
                verdaux: Verdaux = next(verdef.veraux_table.__iter__())
                return verdaux.name, dynent_editor.read_soname()

        for verneed in self.read_version_requirement():
            verneed: Verneed
            for vernaux in verneed.veraux_table:
                vernaux: Vernaux
                if vernaux.other == idx:
                    return vernaux.name, verneed.file

        raise KeyError

    def __str_version_definition(self):
        verdef_table = self.read_version_definition()
        string = '=======\nSection .gnu.version_d\n-------\n'
        for verdef in verdef_table:
            verdef: Verneed
            string += str(verdef) + '\n'
            if not verdef.veraux_table:
                continue
            for verdaux in verdef.veraux_table:
                string += str(verdaux) + '\n'
        return string

    def __str_version_requirement(self):
        verneed_table = self.read_version_requirement()
        string = '=======\nSection .gnu.version_r\n-------\n'
        for verneed in verneed_table:
            verneed: Verneed
            string += str(verneed) + '\n'
            if not verneed.veraux_table:
                continue
            for vernaux in verneed.veraux_table:
                string += str(vernaux) + '\n'
        return string

    def __str__(self):
        string = ''
        string += self.__str_version_definition()
        string += self.__str_version_requirement()

        # string += str(self.dynent_editor)
        return string



from typing import List


def transaction(func):
    def wrapper(self: Transaction, *args, **kwargs):
        self.enqueue_command(func, *args, **kwargs)
    return wrapper


class Transaction:
    class Command:
        func: callable
        args: list
        kwargs: dict

    cmdq: List[Command]

    def __init__(self):
        self.cmdq = []

    def enqueue_command(self, func: callable, *args, **kwargs):
        cmd = Transaction.Command()
        cmd.func = func
        cmd.args = args
        cmd.kwargs = kwargs

        self.cmdq.append(cmd)

    def on_pre_commit(self):
        pass

    def on_post_commit(self):
        pass

    def commit(self):
        self.on_pre_commit()

        for cmd in self.cmdq:
            cmd.func(self, *cmd.args, **cmd.kwargs)

        self.on_post_commit()



from typing import Union

from woodelf import ELF_CLASS, ELF_DATA, ELF_VERSION, NotAnElfError


class E_Ident:
    magic: bytes
    cls: Union[ELF_CLASS, int]
    data: Union[ELF_DATA, int]
    version: Union[ELF_VERSION, int]
    padding: bytes

    def __init__(self, magic: bytes, cls: ELF_CLASS, data: ELF_DATA, version: ELF_VERSION, padding: bytes):
        self.magic = magic
        self.cls = cls
        self.data = data
        self.version = version
        self.padding = padding

    @classmethod
    def size(cls):
        return 16

    @classmethod
    def from_bytes(cls, b: bytes):
        if len(b) < cls.size():
            raise NotAnElfError('file is too small')

        assert len(b) == cls.size()

        ei_magic = b[0:4]
        ei_cls = b[4]
        ei_data = b[5]
        ei_version = b[6]
        ei_pad = b[7:]

        if ei_magic != '\x7fELF'.encode('ascii'):
            raise NotAnElfError('magic field does not match')

        return E_Ident(ei_magic,
                       ELF_CLASS(int(ei_cls)),
                       ELF_DATA(int(ei_data)),
                       ELF_VERSION(int(ei_version)),
                       ei_pad)

    def to_bytes(self):
        b = self.magic + bytes([int(self.cls), int(self.data), int(self.version)]) + self.padding

        assert len(b) == self.size()

        return b

    def __str__(self):
        string = 'E_IDENT{'
        string += 'magic: ' + self.magic.decode('ascii', errors='ignore') + ', '
        string += 'class: ' + str(self.cls) + ', '
        string += 'data: ' + str(self.data) + ', '
        string += 'version: ' + str(self.version) + ', '
        string += 'padding: ' + str(self.padding) + '}'
        return string



from typing import List, Union

from .. import Elf, ELF32, ELF64, EDITOR, SECTION, StrTabEditor, api, DynamicEntryEditor, SYMBOL_BIND, SYMBOL_TYPE, \
    SYMBOL_VISIBILITY
from ..core import Element


class Symbol(Element, api.Symbol):
    @classmethod
    def units(cls, elf: Elf) -> List[Union[ELF32, ELF64]]:
        if elf.unit == ELF32:
            return [elf.unit.Word, elf.unit.Addr, elf.unit.Word, elf.unit.uchar,
                    elf.unit.uchar, elf.unit.Half]
        elif elf.unit == ELF64:
            return [elf.unit.Word, elf.unit.uchar, elf.unit.uchar, elf.unit.Half,
                    elf.unit.Addr, elf.unit.Xword]
        else:
            raise AssertionError

    @classmethod
    def from_bytes(cls, elf: Elf, b: bytes):
        if elf.unit == ELF32:
            st_name, st_value, st_size, st_info, \
            st_other, st_shndx = cls.deserialize(elf, b)
        elif elf.unit == ELF64:
            st_name, st_info, st_other, st_shndx, \
            st_value, st_size = cls.deserialize(elf, b)
        else:
            raise AssertionError

        dynsym_editor: StrTabEditor = elf.get_editor(EDITOR.STRTAB, SECTION.DYNSTR)
        # dynent_editor: DynamicEntryEditor = elf.get_editor(EDITOR.DYNAMIC_ENTRY)
        # dyntab = dynent_editor.read_dynamic_entries()

        s = Symbol()
        s.name = dynsym_editor.get_str(st_name)
        s.value = st_value
        s.siz = st_size
        s.bind = SYMBOL_BIND(cls.__st_bind(st_info))
        s.typ = SYMBOL_TYPE(cls.__st_type(st_info))
        s.other = SYMBOL_VISIBILITY(st_other)
        s.shndx = st_shndx

        return s

    def to_bytes(self, elf: Elf) -> bytes:
        dynsym_editor: StrTabEditor = elf.get_editor(EDITOR.STRTAB, SECTION.DYNSTR)
        # dynent_editor: DynamicEntryEditor = elf.get_editor(EDITOR.DYNAMIC_ENTRY)
        # dyntab = dynent_editor.read_dynamic_entries()

        st_name = dynsym_editor.find(self.name)
        # st_info = -1
        # for i, de in enumerate(dyntab):
        #     if de.tag == self.info.tag and de.un == self.info.un:
        #         st_info = i
        #         break
        # if st_info < 0:
        #     raise AssertionError

        st_info = self.__st_info(int(self.bind), int(self.typ))

        if elf.unit == ELF32:
            return self.serialize(elf, st_name, self.value, self.siz, st_info,
                                  int(self.other), self.shndx)
        elif elf.unit == ELF64:
            return self.serialize(elf, st_name, st_info, int(self.other), self.shndx,
                                  self.value, self.siz)
        raise AssertionError

    def is_defined(self):
        return self.siz != 0 and self.shndx != 0

    def is_needed(self):
        return self.value == 0 and self.siz == 0 and self.shndx == 0

    @classmethod
    def __st_bind(cls, st_info: int) -> int:
        return st_info >> 4

    @classmethod
    def __st_type(cls, st_info: int) -> int:
        return st_info & 0xf

    @classmethod
    def __st_info(cls, st_bind: int, st_type: int) -> int:
        return (st_bind << 4) + (st_type & 0xf)

    def __str__(self):
        string = 'Symbol: '
        string += 'name: ' + str(self.name) + ', '
        string += 'value: ' + hex(self.value) + ', '
        string += 'size: ' + hex(self.siz) + ', '
        string += 'bind: ' + str(self.bind) + ', '
        string += 'type: ' + str(self.typ) + ', '
        string += 'other: ' + str(self.other) + ', '
        string += 'shndx: ' + hex(self.shndx) + '.'
        return string


class SymbolTable(api.SymbolTable):
    def defined_symbols(self):
        return filter(lambda e: e.is_defined(), self)

    def needed_symbols(self):
        return filter(lambda e: e.is_needed(), self)

    def __str__(self):
        string = 'SymbolTable:\n'
        for symbol in self:
            string += str(symbol) + '\n'

        return string



from .. import api, Section
from ..core import Editor, Elf
from ..constants import SECTION


class StrTabEditor(Editor, api.StrTabEditor):
    elf: Elf
    section_type: SECTION
    section: Section

    def __init__(self, elf: Elf, section: SECTION, base_editor: api.Editor = None, _unsafe=False):
        super().__init__(elf)

        self.elf = elf
        self.section_type = section
        self.section = elf.get_section(section, _unsafe=_unsafe)

    def append(self, string: str):
        bstring = string.encode('ascii') + b'\0'
        content = self.section.read_content() + bstring
        self.section.write_content(content)

    def find(self, string: str) -> int:
        bstring = string.encode('ascii') + b'\0'
        return self.section.find(bstring)

    def has(self, string: str) -> bool:
        return self.find(string) >= 0

    def get_str(self, pos: int) -> str:
        c = self.section.read_content()
        end = c.find(b'\0', pos)

        return c[pos:end].decode(encoding='ascii')

    def __str__(self):
        string = '=======\nSection ' + self.section.name + '\n-------\n'
        string += self.section.read_content()
        return string



from typing import List, Union

from .. import api, Elf, ELF32, ELF64, SEGMENT_TYPE
from ..core import Element


class ProgramHeader(api.ProgramHeader, Element):
    @classmethod
    def units(cls, elf: Elf) -> List[Union[ELF32, ELF64]]:
        if elf.unit == ELF32:
            return [elf.unit.Word, elf.unit.Off, elf.unit.Addr, elf.unit.Addr,
                    elf.unit.Word, elf.unit.Word, elf.unit.Word, elf.unit.Word]
        elif elf.unit == ELF64:
            return [elf.unit.Word, elf.unit.Word, elf.unit.Off, elf.unit.Addr,
                    elf.unit.Addr, elf.unit.Xword, elf.unit.Xword, elf.unit.Xword]
        else:
            raise AssertionError

    @classmethod
    def from_bytes(cls, elf: Elf, b: bytes):
        if elf.unit == ELF32:
            p_type, p_offset, p_vaddr, p_paddr, \
            p_filesz, p_memsz, p_flags, p_align \
                = cls.deserialize(elf, b)
        elif elf.unit == ELF64:
            p_type, p_flags, p_offset, p_vaddr, \
            p_paddr, p_filesz, p_memsz, p_align \
                = cls.deserialize(elf, b)
        else:
            raise AssertionError

        ph = ProgramHeader()
        try:
            ph.type = SEGMENT_TYPE(p_type)
        except ValueError:
            ph.type = p_type
        ph.offset = p_offset
        ph.vaddr = p_vaddr
        ph.paddr = p_paddr
        ph.filesz = p_filesz
        ph.memsz = p_memsz
        ph.flags = p_flags
        ph.align = p_align

        return ph

    def to_bytes(self, elf: Elf) -> bytes:
        if elf.unit == ELF32:
            return self.serialize(elf, int(self.type), self.offset, self.vaddr, self.paddr,
                                  self.filesz, self.memsz, self.flags, self.align)
        elif elf.unit == ELF64:
            return self.serialize(elf, int(self.type), self.flags, self.offset, self.vaddr,
                                  self.paddr, self.filesz, self.memsz, self.align)
        else:
            raise AssertionError

    def __str__(self):
        string = 'Program header (segment): '
        string += 'type: ' + str(self.type) + ', '
        string += 'offset: ' + hex(self.offset) + ', '
        string += 'vaddr: ' + hex(self.vaddr) + ', '
        string += 'paddr: ' + hex(self.paddr) + ', '
        string += 'filesz: ' + hex(self.filesz) + ', '
        string += 'memsz: ' + hex(self.memsz) + ', '
        string += 'flags: ' + hex(self.flags) + ', '
        string += 'align: ' + hex(self.align) + '.'

        return string


class ProgramHeaderTable(api.ProgramHeaderTable):
    def __str__(self):
        string = 'Program Header (Segment) Table: \n'
        for ph in self:
            string += str(ph) + '\n'
        return string



from enum import Enum, auto


class IntEnum(Enum):
    def __int__(self):
        return self.value


class ELF_CLASS(IntEnum):
    CLASSNONE = 0
    CLASS32 = 1
    CLASS64 = 2

    # @classmethod
    # def _missing_(cls, value):
    #     return value


class ELF_DATA(IntEnum):
    DATANONE = 0
    DATA2LSB = 1
    DATA2MSB = 2

    # @classmethod
    # def _missing_(cls, value):
    #     return value

    def __str__(self):
        if self == ELF_DATA.DATA2LSB:
            return 'little'
        elif self == ELF_DATA.DATA2MSB:
            return 'big'
        else:
            return 'unknown'


class ELF_TYPE(IntEnum):
    NONE = 0
    REL = 1
    EXEC = 2
    DYN = 3
    CORE = 4

    # @classmethod
    # def _missing_(cls, value):
    #     return value


class ELF_MACHINE(IntEnum):
    EM_NONE = 0
    EM_M32 = 1
    EM_SPARC = 2
    EM_386 = 3
    EM_68K = 4
    EM_88K = 5
    EM_860 = 7
    EM_MIPS = 8
    EM_ARM = 40
    EM_X86_64 = 62
    EM_AARCH64 = 183
    UNKNOWN = -1

    # @classmethod
    # def _missing_(cls, value):
    #     return value


class ELF_VERSION(IntEnum):
    NONE = 0
    CURRENT = 1

    # @classmethod
    # def _missing_(cls, value):
    #     return value

# https://docs.oracle.com/cd/E19957-01/806-0641/6j9vuqujo/index.html


class ELF32(IntEnum):
    Addr = 4
    Half = 2
    Off = 4
    Sword = 4
    Word = 4
    uchar = 1

    # 200830: Xword and Sxword are added for convenience
    Xword = 4
    Sxword = 4


class ELF64(IntEnum):
    Addr = 8
    Half = 2
    Off = 8
    Sword = 4
    Word = 4
    Xword = 8
    Sxword = 8
    uchar = 1


class EDITOR(Enum):
    SYMBOL_VERSION = auto()
    DYNAMIC_ENTRY = auto()
    STRTAB = auto()
    ELF_HEADER = auto()
    SECTION_HEADER = auto()
    PROGRAM_HEADER = auto()
    SYMBOL = auto()


class SEGMENT_TYPE(IntEnum):
    NULL = 0
    LOAD = 1
    DYNAMIC = 2
    INTERP = 3
    NOTE = 4
    SHLIB = 5
    PHDR = 6

    EH_FRAME = 0x6474e550
    STACK = 0x6474e551
    RELRO = 0x6474e552

    # @classmethod
    # def _missing_(cls, value):
    #     return None


class SECTION(Enum):
    NOTE_ANDROID_IDENT = '.note.android.ident'
    NOTE_GNU_BUILD_ID = '.note.gnu.build-id'
    DYNSYM = '.dynsym'
    DYNSTR = '.dynstr'
    GNU_HASH = '.gnu.hash'
    GNU_VERSION = '.gnu.version'
    GNU_VERSION_D = '.gnu.version_d'
    GNU_VERSION_R = '.gnu.version_r'
    RELA_DYN = '.rela.dyn'
    RELA_PLT = '.rela.plt'
    PLT = '.plt'
    TEXT = '.text'
    RODATA = '.rodata'
    GCC_EXCEPT_TABLE = '.gcc_except_table'
    EH_FRAME = '.eh_frame'
    EH_FRAME_HDR = '.eh_frame_hdr'
    FINI_ARRAY = '.fini_array'
    INIT_ARRAY = '.init_array'
    DATA_REL_RO = '.data.rel.ro'
    DYNAMIC = '.dynamic'
    GOT = '.got'
    GOT_PLT = '.got.plt'
    DATA = '.data'
    BSS = '.bss'
    NOTE_GNU_GOLD_VERSION = '.note.gnu.gold-version'
    GNU_DEBUGDATA = '.gnu_debugdata'
    SHSTRTAB = '.shstrtab'

    def __str__(self):
        return self.value


class DYNAMIC_ENTRY_TAG(Enum):
    DT_NULL = 0
    DT_NEEDED = 1
    DT_PLTRELSZ = 2
    DT_PLTGOT = 3
    DT_HASH = 4
    DT_STRTAB = 5
    DT_SYMTAB = 6
    DT_RELA = 7
    DT_RELASZ = 8
    DT_RELAENT = 9
    DT_STRSZ = 10
    DT_SYMENT = 11
    DT_INIT = 12
    DT_FINI = 13
    DT_SONAME = 14
    DT_RPATH = 15
    DT_SYMBOLIC = 16
    DT_REL = 17
    DT_RELSZ = 18
    DT_RELENT = 19
    DT_PLTREL = 20
    DT_DEBUG = 21
    DT_TEXTREL = 22
    DT_JMPREL = 23

    #/ *http: // www.sco.com / developers / gabi / latest / ch5.dynamic.html * /
    DT_BIND_NOW = 24
    DT_INIT_ARRAY = 25
    DT_FINI_ARRAY = 26
    DT_INIT_ARRAYSZ = 27
    DT_FINI_ARRAYSZ = 28
    DT_RUNPATH = 29
    DT_FLAGS = 30
    #/ *glibc and BSD disagree for DT_ENCODING; glibc looks wrong.* /
    DT_PREINIT_ARRAY = 32
    DT_PREINIT_ARRAYSZ = 33

    #/ *Experimental support for SHT_RELR sections.For details, see proposal
    # at https: // groups.google.com / forum /  # !topic/generic-abi/bX460iggiKg */
    DT_RELR = 0x6fffe000
    DT_RELRSZ = 0x6fffe001
    DT_RELRENT = 0x6fffe003
    DT_RELRCOUNT = 0x6fffe005

    # bionic/libc/include/elf.h
    DT_LOOS = 0x6000000d
    DT_ANDROID_REL = DT_LOOS + 2
    DT_ANDROID_RELSZ = DT_LOOS + 3
    DT_ANDROID_RELA = DT_LOOS + 4
    DT_ANDROID_RELASZ = DT_LOOS + 5

    DT_GNU_HASH = 0x6ffffef5
    DT_TLSDESC_PLT = 0x6ffffef6
    DT_TLSDESC_GOT = 0x6ffffef7

    # bionic/libc/kernel/uapi/linux/elf.h
    DT_VERSYM = 0x6ffffff0
    DT_RELACOUNT = 0x6ffffff9
    DT_RELCOUNT = 0x6ffffffa
    DT_FLAGS_1 = 0x6ffffffb
    DT_VERDEF = 0x6ffffffc
    DT_VERDEFNUM = 0x6ffffffd
    DT_VERNEED = 0x6ffffffe
    DT_VERNEEDNUM = 0x6fffffff

    DT_UNKNOWN = None

    def __int__(self):
        if self == DYNAMIC_ENTRY_TAG:
            return 0

        return self.value


STR_DT = [DYNAMIC_ENTRY_TAG.DT_NEEDED,
          DYNAMIC_ENTRY_TAG.DT_SONAME,
          DYNAMIC_ENTRY_TAG.DT_RPATH,
          DYNAMIC_ENTRY_TAG.DT_RUNPATH]

PTR_DT = [DYNAMIC_ENTRY_TAG.DT_PLTGOT,
          DYNAMIC_ENTRY_TAG.DT_HASH,
          DYNAMIC_ENTRY_TAG.DT_STRTAB,
          DYNAMIC_ENTRY_TAG.DT_SYMTAB,
          DYNAMIC_ENTRY_TAG.DT_RELA,
          DYNAMIC_ENTRY_TAG.DT_INIT,
          DYNAMIC_ENTRY_TAG.DT_FINI,
          DYNAMIC_ENTRY_TAG.DT_REL,
          DYNAMIC_ENTRY_TAG.DT_DEBUG,
          DYNAMIC_ENTRY_TAG.DT_JMPREL,
          DYNAMIC_ENTRY_TAG.DT_INIT_ARRAY,
          DYNAMIC_ENTRY_TAG.DT_FINI_ARRAY,
          DYNAMIC_ENTRY_TAG.DT_PREINIT_ARRAY,
          DYNAMIC_ENTRY_TAG.DT_ANDROID_REL,
          DYNAMIC_ENTRY_TAG.DT_ANDROID_RELA,
          DYNAMIC_ENTRY_TAG.DT_VERSYM,
          DYNAMIC_ENTRY_TAG.DT_VERDEF,
          DYNAMIC_ENTRY_TAG.DT_VERNEED,
          DYNAMIC_ENTRY_TAG.DT_GNU_HASH]


class SYMBOL_BIND(IntEnum):
    STB_LOCAL = 0
    STB_GLOBAL = 1
    STB_WEAK = 2


class SYMBOL_TYPE(IntEnum):
    STT_NOTYPE = 0
    STT_OBJECT = 1
    STT_FUNC = 2
    STT_SECTION = 3
    STT_FILE = 4
    STT_COMMON = 5

    STT_SPARC_REGISTER = 13


class SYMBOL_VISIBILITY(IntEnum):
    STV_DEFAULT = 0
    STV_INTERNAL = 1
    STV_HIDDEN = 2
    STV_PROTECTED = 3



from .. import api
from ..core import Editor
from ..elements import E_Ident, ElfHeader


class ElfHeaderEditor(api.ElfHeaderEditor, Editor):

    def read_e_ident(self) -> E_Ident:
        rev = self.elf.get_current_revision()
        cache = self.elf.get_cache(rev, 'e_ident')

        if e_ident := cache.lookup():
            return e_ident

        with open(rev, 'rb') as f:
            b = f.read(E_Ident.size())

        e_ident = E_Ident.from_bytes(b)
        cache.update(e_ident)
        return e_ident

    def read_elf_header(self, rev_idx: int = -1) -> ElfHeader:
        rev = self.elf.revisions[rev_idx]
        cache = self.elf.get_cache(rev, 'elfh')

        if elfh := cache.lookup():
            return elfh

        with open(rev, 'rb') as f:
            b = f.read(ElfHeader.size(self.elf))

        elfh = ElfHeader.from_bytes(self.elf, b)
        cache.update(elfh)
        return elfh

    def write_elf_header(self, elf_header: ElfHeader):
        rev = self.elf.get_current_revision()

        with open(rev, 'r+b') as f:
            f.write(elf_header.to_bytes(self.elf))

        self.elf.get_cache(rev, 'e_ident').invalidate()
        self.elf.get_cache(rev, 'elfh').invalidate()




from .dynamic_entry import DynamicEntry
from .e_ident import E_Ident
from .elf_header import ElfHeader
from .symbol_version import Verdef, Verdaux, Verneed, Vernaux, Version, \
    VerauxTable, VerdefTable, VerneedTable, VersionTable
from .symbol import SymbolTable, Symbol



import setuptools
import os

# borrowed from https://stackoverflow.com/a/36693250
def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


deps = [
    'sh',
    'hexdump',
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="woodelf", # Replace with your own username
    version="0.1.0",
    author="Donghwi Kim",
    author_email="dhkim09@kaist.ac.kr",
    description="Pure Python ELF file parser",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # url="https://github.com/dhkim09a/subcmdparse",
    packages=setuptools.find_packages(where='src'),
    # classifiers=[
    #     "Programming Language :: Python :: 3",
    #     "License :: OSI Approved :: MIT License",
    #     "Operating System :: OS Independent",
    # ],
    install_requires=deps,
    extras_require={
        'test': deps,
    },
    package_dir={
        '': 'src',
    },
    # package_data={'': package_files('yautil/mountutil/yaffs2utils-0.2.9')},
    python_requires='>=3.9',
)



from typing import List, Union

from .. import Elf, ELF32, ELF64, api
from ..core import Element


class GnuHash(Element, api.GnuHash):
    @classmethod
    def units(cls, elf: Elf) -> List[Union[ELF32, ELF64]]:
        return [elf.unit.Word, elf.unit.Word, elf.unit.Word, elf.unit.Word]

    @classmethod
    def from_bytes(cls, elf: Elf, b: bytes):
        nbuckets, symndx, maskwords, shift2 = cls.deserialize(elf, b[:int(elf.unit.Word) * 4])

        gnu_hash = GnuHash()
        gnu_hash.nbuckets = nbuckets
        gnu_hash.symndx = symndx
        gnu_hash.maskwords = maskwords
        gnu_hash.shift2 = shift2

        # FIXME: This is incomplete
        # https://blogs.oracle.com/solaris/gnu-hash-elf-sections-v2
        # https://www.gabriel.urdhr.fr/2015/09/28/elf-file-format/#gnu-hash-table


    def to_bytes(self, elf: Elf) -> bytes:
        pass


from .dynamic_entry_editor import DynamicEntryEditor
from .strtab_editor import StrTabEditor
from .symbol_version_editor import SymbolVersionEditor
from .elf_header_editor import ElfHeaderEditor
from .section_header_editor import SectionHeaderEditor
from .program_header_editor import ProgramHeaderEditor
from .symbol_editor import SymbolEditor


from typing import Union, List

from .. import EDITOR, SECTION, api, ELF64, ELF32, gnu_hash
from ..api import Elf, SymbolVersionEditor, StrTabEditor

from ..core.element import Element


class Veraux(Element):
    next: object

    def __init__(self):
        self.next = None


class VerauxTable:
    head: Union[Veraux, None]
    size: int

    class __Iterator:
        next: Union[Veraux, None]

        def __init__(self, head: Union[Veraux, None]):
            self.next = head

        def __next__(self) -> Veraux:
            if self.next:
                cur = self.next
                if self.next.next:
                    assert isinstance(self.next.next, Veraux)
                    self.next = self.next.next
                else:
                    self.next = None
                return cur
            else:
                raise StopIteration

    def __init__(self):
        self.head = None
        self.size = 0

    def __iter__(self):
        return VerauxTable.__Iterator(self.head)

    def __len__(self):
        return self.size

    def __getitem__(self, key: int):
        if not isinstance(key, int):
            raise TypeError('list indices must be integers')

        if not (v := self.head) or not (0 <= key <= len(self)):
            raise IndexError('list index out of range')

        if key == 0:
            return self.head
        else:
            while (next_v := v.next) and (key := key - 1):
                v = next_v
            if not next_v:
                raise IndexError('list index out of range')
            return next_v

    def __setitem__(self, key: int, value: Veraux):
        if not isinstance(key, int):
            raise TypeError('list indices must be integers')

        self.insert(key, value)

    def append(self, version: Veraux):
        self.size += 1

        if not (v := self.head):
            self.head = version
            return

        while next_v := v.next:
            v = next_v

        v.next = version

    def insert(self, index: int, version: Veraux):
        if not isinstance(version, Veraux):
            raise ValueError

        self.size += 1

        if not (v := self.head):
            self.head = version
            return

        if index > 0 and index % len(self) == 0:
            index = len(self)
        else:
            index = index % len(self)

        assert 0 <= index <= len(self)

        if index == 0:
            version.next = self.head
            self.head = version
        else:
            while (next_v := v.next) and (index := index - 1):
                v = next_v
            v.next = version
            version.next = next_v


class Ver(Veraux):
    cnt: int
    aux: Union[Veraux, None]

    veraux_table: Union[VerauxTable, None]

    def __init__(self):
        super().__init__()
        self.veraux_table = None

    def append_veraux(self, veraux: Veraux):
        if not self.veraux_table:
            self.veraux_table = VerauxTable()
            self.aux = veraux

        self.veraux_table.append(veraux)

        if self.cnt < len(self.veraux_table):
            self.cnt = len(self.veraux_table)

    def insert_veraux(self, index: int, veraux: Veraux):
        if not self.veraux_table:
            self.veraux_table = VerauxTable()
            self.aux = veraux

        self.veraux_table.insert(index, veraux)

        if self.cnt < len(self.veraux_table):
            self.cnt = len(self.veraux_table)


class Version(api.Version, Element):
    @classmethod
    def units(cls, elf: Elf) -> List[Union[ELF32, ELF64]]:
        # This must be the same with vna_other
        return [elf.unit.Half]

    @classmethod
    def from_bytes(cls, elf: Elf, b: bytes):
        symver_editor: SymbolVersionEditor = elf.get_editor(EDITOR.SYMBOL_VERSION)
        ver = Version()
        vna_other = cls.deserialize(elf, b)
        if vna_other != 0:
            name, soname = symver_editor.get_vername_soname_by_version(vna_other)
            ver.name = name
            ver.soname = soname
        else:
            ver.name = ver.soname = None

        return ver

    def to_bytes(self, elf: Elf) -> bytes:
        symver_editor: SymbolVersionEditor = elf.get_editor(EDITOR.SYMBOL_VERSION)
        if self.name:
            value = symver_editor.get_version_by_name(self.name, self.soname)
        else:
            value = 0
        return self.serialize(elf, value)

    def is_local(self):
        return not self.name

    def __str__(self):
        if self.is_local():
            return 'LOCAL'
        string = self.name
        if self.soname:
            string += ' (from ' + self.soname + ')'
        return string


class Verdaux(Veraux, api.Verdaux):
    name: str
    next: Union[Veraux, None]

    def __init__(self, vda_name: str):
        super().__init__()
        self.name = vda_name

    @classmethod
    def size(cls, elf: Elf) -> int:
        return int(elf.unit.Word) * 2

    @classmethod
    def from_bytes(cls, elf: Elf, b: bytes):
        assert len(b) == Verdaux.size(elf)

        pos = 0
        vda_name = int.from_bytes(b[pos:(pos := pos + int(elf.unit.Word))], byteorder=elf.endian, signed=False)
        vda_next = int.from_bytes(b[pos:(pos := pos + int(elf.unit.Word))], byteorder=elf.endian, signed=False)

        if vda_next != 0 and vda_next != Verdaux.size(elf):
            print('error: weird vn_aux offset')

        name = elf.get_editor(EDITOR.STRTAB, SECTION.DYNSTR).get_str(vda_name)

        verdaux = Verdaux(name)

        return verdaux

    def to_bytes(self, elf: Elf):
        # dynstr_sec: lief.ELF.Section = elf.get_section('.dynstr')
        if (pos := elf.get_editor(EDITOR.STRTAB, SECTION.DYNSTR).find(self.name)) >= 0:
            b = pos.to_bytes(int(elf.unit.Word), byteorder=elf.endian, signed=False)
        else:
            b = (0).to_bytes(int(elf.unit.Word), byteorder=elf.endian, signed=False)
        if self.next:
            b += (Verdaux.size(elf)).to_bytes(int(elf.unit.Word), byteorder=elf.endian, signed=False)
        else:
            b += (0).to_bytes(int(elf.unit.Word), byteorder=elf.endian, signed=False)

        assert len(b) == Verdaux.size(elf)

        return b

    def __str__(self):
        return '(verdaux) name: ' + str(self.name) \
               + ' (gnu hash: ' + hex(gnu_hash(self.name)) + ')'


class Verdef(Ver, api.Verdef):
    version: int
    flags: int
    ndx: int
    cnt: int
    hash: int
    aux: Union[Verdaux, None]
    next: Union[Veraux, None]

    def __init__(self, vda_name: str = None, vd_ndx: int = 1):
        super().__init__()

        self.version = 1
        self.flags = 0
        self.ndx = vd_ndx
        self.cnt = 0

        if vda_name:
            self.hash = gnu_hash(vda_name)
            verdaux = Verdaux(vda_name)
            self.append_veraux(verdaux)
        else:
            self.hash = 0

    @classmethod
    def units(cls, elf: Elf) -> List[Union[ELF32, ELF64]]:
        return [elf.unit.Half, elf.unit.Half, elf.unit.Half, elf.unit.Half,
                elf.unit.Word, elf.unit.Word, elf.unit.Word]

    @classmethod
    def from_bytes(cls, elf: Elf, b: bytes):

        vd_version, vd_flags, vd_ndx, vd_cnt, vd_hash, vd_aux, vd_next \
            = cls.deserialize(elf, b)

        if vd_version != 1:
            print('error: vd_version must be 1')
        if vd_aux != 0 and vd_aux != Verdef.size(elf):
            print('error: weird vd_aux offset')

        verdef = Verdef()

        verdef.version = vd_version
        verdef.flags = vd_flags
        verdef.ndx = vd_ndx
        verdef.cnt = vd_cnt
        verdef.hash = vd_hash

        return verdef

    def to_bytes(self, elf: Elf):
        if self.aux:
            aux = Verdef.size(elf)
        else:
            aux = 0
        if self.next:
            nxt = Verdef.size(elf) + self.cnt * Verdaux.size(elf)
        else:
            nxt = 0

        return self.serialize(elf, self.version, self.flags, self.ndx, self.cnt, self.hash, aux, nxt)

    def __str__(self):
        return '(verdef) version: ' + str(self.version) \
               + ', flags: ' + str(self.flags) \
               + ', ndx: ' + str(self.ndx) \
               + ', cnt: ' + str(self.cnt) \
               + ', hash: ' + hex(self.hash)

    def update(self, new_version: str, ndx: int):
        self.hash = gnu_hash(new_version)
        self.ndx = ndx
        self.insert_veraux(0, Verdaux(new_version))

    def get_ndx(self):
        return self.ndx


class Vernaux(Veraux, api.Vernaux):
    hash: int
    flags: int
    next: Union[Veraux, None]

    def __init__(self, vna_name: str):
        super().__init__()

        # constants
        self.flags = 0

        # init from params
        self.name = vna_name
        self.other = 0
        self.hash = gnu_hash(vna_name)

    @classmethod
    def units(cls, elf: Elf) -> List[Union[ELF32, ELF64]]:
        return [elf.unit.Word, elf.unit.Half, elf.unit.Half, elf.unit.Word,
                elf.unit.Word]

    @classmethod
    def from_bytes(cls, elf: Elf, b: bytes):
        vna_hash, vna_flags, vna_other, vna_name, \
        vna_next = cls.deserialize(elf, b)

        if vna_next != 0 and vna_next != Vernaux.size(elf):
            print('error: weird vn_aux offset')

        name = elf.get_editor(EDITOR.STRTAB, SECTION.DYNSTR).get_str(vna_name)

        vernaux = Vernaux(name)

        vernaux.other = vna_other
        vernaux.hash = vna_hash
        vernaux.flags = vna_flags

        return vernaux

    def __str__(self):
        return '(vernaux) hash: ' + hex(self.hash) \
               + ', flags: ' + str(self.flags) \
               + ', other: ' + str(self.other) \
               + ', name: ' + str(self.name) \
               + ' (gnu hash: ' + hex(gnu_hash(self.name)) + ')'

    def to_bytes(self, elf: Elf):
        dynstr_edit: StrTabEditor = elf.get_editor(EDITOR.STRTAB, SECTION.DYNSTR)
        symver_edit: SymbolVersionEditor = elf.get_editor(EDITOR.SYMBOL_VERSION)

        if (vna_name := dynstr_edit.find(self.name)) < 0:
            vna_name = 0

        if self.next:
            vna_next = Vernaux.size(elf)
        else:
            vna_next = 0

        return self.serialize(elf, self.hash, self.flags, self.other, vna_name, vna_next)


class Verneed(Ver, api.Verneed):
    version: int
    cnt: int
    file: str
    aux: Union[Vernaux, None]
    next: Union[Veraux, None]

    @classmethod
    def size(cls, elf: Elf) -> int:
        return int(elf.unit.Half) * 2 + int(elf.unit.Word) * 3

    @classmethod
    def from_bytes(cls, elf: Elf, b: bytes):
        assert len(b) == Verneed.size(elf)

        pos = 0

        vn_version = int.from_bytes(b[pos:(pos := pos + int(elf.unit.Half))], byteorder=elf.endian, signed=False)
        vn_cnt = int.from_bytes(b[pos:(pos := pos + int(elf.unit.Half))], byteorder=elf.endian, signed=False)
        vn_file = int.from_bytes(b[pos:(pos := pos + int(elf.unit.Word))], byteorder=elf.endian, signed=False)
        vn_aux = int.from_bytes(b[pos:(pos := pos + int(elf.unit.Word))], byteorder=elf.endian, signed=False)
        # vn_next = int.from_bytes(b[pos:(pos := pos + int(elf.unit.Word))], byteorder=elf.endian, signed=False)

        if vn_version != 1:
            print('error: vn_version must be 1')
        if vn_aux != 0 and vn_aux != Verneed.size(elf):
            print('error: weird vn_aux offset')

        file = elf.get_editor(EDITOR.STRTAB, SECTION.DYNSTR).get_str(vn_file)

        verneed = Verneed(file)

        verneed.version = vn_version
        verneed.cnt = vn_cnt

        return verneed

    def __init__(self, vn_file: str):
        super().__init__()

        # constants
        self.version = 1

        # init from params
        self.file = vn_file

        # just init
        self.cnt = 0
        self.aux = None

    def __str__(self):
        return '(verneed) version: ' + str(self.version) \
               + ', cnt: ' + str(self.cnt) \
               + ', file: ' + str(self.file)

    def to_bytes(self, elf: Elf):
        b = self.version.to_bytes(int(elf.unit.Half), byteorder=elf.endian, signed=False)
        b += self.cnt.to_bytes(int(elf.unit.Half), byteorder=elf.endian, signed=False)
        # dynstr_sec: lief.ELF.Section = elf.get_section('.dynstr')
        if (pos := elf.get_editor(EDITOR.STRTAB, SECTION.DYNSTR).find(self.file)) >= 0:
            b += pos.to_bytes(int(elf.unit.Word), byteorder=elf.endian, signed=False)
        else:
            b += (0).to_bytes(int(elf.unit.Word), byteorder=elf.endian, signed=False)
        if self.aux:
            b += (Verneed.size(elf)).to_bytes(int(elf.unit.Word), byteorder=elf.endian, signed=False)
        else:
            b += (0).to_bytes(4, byteorder=elf.endian, signed=False)
        if self.next:
            b += (Verneed.size(elf) + self.cnt * Vernaux.size(elf)).to_bytes(int(elf.unit.Word), byteorder=elf.endian,
                                                                             signed=False)
        else:
            b += (0).to_bytes(int(elf.unit.Word), byteorder=elf.endian, signed=False)

        assert len(b) == Verneed.size(elf)

        return b


class VersionTable(api.VersionTable):
    pass


class VerdefTable(VerauxTable, api.VerdefTable):
    head: Union[Verdef, None]
    next: Union[Verdef, None]

    def add_entry(self, vda_name: str) -> Verdef:
        vd_ndx = self[len(self) - 1].ndx + 1
        if matches := [e for e in filter(lambda e: e.ndx == vd_ndx, self)]:
            assert len(matches) == 1
            verdef: Verdef = matches[0]
            verdef.update(vda_name, vd_ndx)
        else:
            verdef = Verdef(vda_name=vda_name, vd_ndx=vd_ndx)
            self.append(verdef)

        return verdef


class VerneedTable(VerauxTable, api.VerneedTable):
    head: Union[Verneed, None]
    next: Union[Verneed, None]

    def add_entry(self, vn_file: str, vna_name: str):
        if matches := [e for e in filter(lambda e: e.file == vn_file, self)]:
            assert len(matches) == 1
            verneed: Verneed = matches[0]
            if matches := [e for e in filter(lambda e: e.name == vna_name, verneed.veraux_table)]:
                assert len(matches) == 1
            else:
                vernaux = Vernaux(vna_name)
                verneed.append_veraux(vernaux)
        else:
            verneed = Verneed(vn_file)
            vernaux = Vernaux(vna_name)
            verneed.append_veraux(vernaux)
            self.append(verneed)



from typing import List

from .. import api
from ..core import Editor, Elf
from ..constants import SECTION, DYNAMIC_ENTRY_TAG
from ..elements import DynamicEntry


class DynamicEntryEditor(Editor, api.DynamicEntryEditor):
    elf: Elf

    def __init__(self, elf: Elf):
        super().__init__(elf)
        self.elf = elf

    def read_dynamic_entries(self, rev_idx: int = -1) -> List[DynamicEntry]:
        dynamic = self.elf.get_section(SECTION.DYNAMIC)
        rev = self.elf.revisions[rev_idx]
        cache = self.elf.get_cache(rev, 'dyn_ents')

        if dynlist := cache.lookup():
            return dynlist

        c = dynamic.read_content(rev_idx=rev_idx)

        dynlist = []
        while c:
            dyn_bytes = c[0:DynamicEntry.size(self.elf)]
            c = c[DynamicEntry.size(self.elf):]
            dyn = DynamicEntry.from_bytes(self.elf, dyn_bytes)
            dynlist.append(dyn)

        cache.update(dynlist)

        return dynlist

    def write_dynamic_entries(self, dynlist: List[DynamicEntry]):
        rev = self.elf.get_current_revision()
        cache = self.elf.get_cache(rev, 'dyn_ents')

        b = bytes()
        for dyn in dynlist:
            b += dyn.to_bytes(self.elf)
        self.elf.get_section(SECTION.DYNAMIC).write_content(b)

        cache.invalidate()

    def read_soname(self, rev_idx: int = -1):
        dyn_ents = self.read_dynamic_entries(rev_idx=rev_idx)
        soname_ents = [e for e in filter(lambda e: e.tag == DYNAMIC_ENTRY_TAG.DT_SONAME, dyn_ents)]

        assert len(soname_ents) == 1

        soname = soname_ents[0].un

        assert isinstance(soname, str)

        return soname

    def __str__(self):
        dynlist = self.read_dynamic_entries()
        string = '=======\nSection .dynamic\n-------\n'
        for dyn in dynlist:
            string += str(dyn) + '\n'
        return string



from typing import List

from .. import api, Elf
from ..constants import SECTION


class Editor(api.Editor):
    elf: Elf

    def __init__(self, elf: Elf):
        self.elf = elf

    # def get_section(self, section: SECTION):
    #     return self.elf.get_section(self, section)
